@@ -4,7 +4,7 @@ use warnings FATAL => 'all';
 
 package namespace::local;
 
-our $VERSION = '0.07_07';
+our $VERSION = '0.07_08';
 
 =head1 NAME
 
@@ -122,17 +122,19 @@ The following symbols are not touched by this module, to avoid breaking things:
 
 =item * Arrays: C<@CARP_NOT>, C<@EXPORT>, C<@EXPORT_OK>, C<@ISA>;
 
-=item * Scalars: C<$AUTOLOAD>, C<$a>, C<$b>;
+=item * Hashes: C<%OVERLOAD>;
 
 =item * Files: C<DATA>, C<STDERR>, C<STDIN>, C<STDOUT>;
 
 =item * Functions: C<AUTOLOAD>, C<DESTROY>, C<import>;
 
+=item * Scalars: C<$AUTOLOAD>, C<$a>, C<$b>;
+
 =back
 
 This list is likely incomplete, and may grow in the future.
 
-=head1 METHOD/FUNCTIONS
+=head1 METHODS/FUNCTIONS
 
 None.
 
@@ -175,7 +177,6 @@ No perfect solution has yet been found.
 
 =cut
 
-# this was stolen from namespace::clean
 use B::Hooks::EndOfScope 'on_scope_end';
 
 my @stack;
@@ -186,7 +187,7 @@ sub import {
     my $command = namespace::local::_command->new( caller => [ caller ] );
     $command->parse_options( @_ );
 
-    # on_scope_end callback execution order is direct
+    # on_scope_end executes multiple callbacks as FIFO
     # we need reversed order, so use a stack of commands.
     $stack[-1]->set_next( $command ) if @stack;
     push @stack, $command;
@@ -201,14 +202,58 @@ sub import {
     };
 };
 
-# Hide internal OO engine
-# Maybe it will be released later...
+=head1 ENVIRONMENT
+
+Set C<PERL_NAMESPACE_LOCAL=debug> to see some debugging information
+upon module load.
+
+=head1 THE INTERNALS
+
+A stack of "command" objects is used behind the scenes.
+
+Its interface is not public, see this module's source.
+
+Calling
+
+=over
+
+=item new
+
+=back
+
+on this package will create a command object,
+I<not> a C<namespace::local> instance.
+
+The creation and destruction of command has no effect on the namespaces.
+
+Instead, special C<prepare> and C<execute> methods
+are called upon import and leaving scope, respectively.
+
+=cut
+
+sub new {
+    my $unused = shift;
+    namespace::local::_command->new(@_);
+};
 
 package
     namespace::local::_command;
 
+# FIRST AND FOREMOST
+
+# See C<perldoc perlmod> for how perl stores the Symbol Tables.
+
+# In this module we use two-level hashrefs to represent the table:
+# $table->{ $name }{ $type } = $reference
+# where $name is a function/variable/whatever name,
+# $type is one of ARRAY, CODE, FORMAT, HASH, IO, and SCALAR (see @TYPES below),
+# and $reference is a reference of corresponding type (or undef).
+
+# So Foo::bar() would be represented as $table->{foo}{CODE}
+# whereas @Foo::ISA is $table->{ISA}{ARRAY}.
+
 use Carp;
-use Scalar::Util qw(refaddr);
+use Scalar::Util qw(blessed refaddr reftype);
 our @CARP_NOT = qw(namespace::local);
 
 # TODO need better env parsing...
@@ -216,11 +261,11 @@ use constant DEBUG => ( lc ($ENV{PERL_NAMESPACE_LOCAL} || '' ) eq 'debug' ? 1 :
 
 ### Setup methods
 
-# requires caller => [caller] argument
 sub new {
     my ($class, %opt) = @_;
 
     # TODO check options
+    $opt{caller}     ||= [ caller 0 ];
     $opt{except_rex} ||= qr/^[0-9]+$|^_$/; # no matter what, exempt $_, $1, ...
     $opt{only_rex}   ||= qr/^/; # match all
     $opt{action}     ||= '-below';
@@ -232,6 +277,7 @@ sub new {
     # NOTE if you change the list, also change the EXEMPTIONS section in the POD.
     $opt{touch_not}{$_}{ARRAY}++  for qw( CARP_NOT EXPORT EXPORT_OK ISA );
     $opt{touch_not}{$_}{CODE}++   for qw( AUTOLOAD DESTROY import );
+    $opt{touch_not}{$_}{HASH}++   for qw( OVERLOAD );
     $opt{touch_not}{$_}{IO}++     for qw( DATA STDERR STDIN STDOUT );
     $opt{touch_not}{$_}{SCALAR}++ for qw( AUTOLOAD a b );
 
@@ -249,17 +295,19 @@ sub set_next {
 
 sub DESTROY {
     my $self = shift;
-    carp "probably a bug in namespace::local: callback set but not executed"
+    carp "probably a bug in namespace::local: callback set at $self->{origin} but never executed"
         if $self->{todo} and !$self->{done};
 };
 
 my %known_action;
 $known_action{$_}++ for qw(-above -below -around);
 
-# this changes nothing except the object itself
+# This changes nothing except the object itself
+# input is the same as that of namespace::local->import
 sub parse_options {
     my $self = shift;
 
+    # wrote a Getopt::Long from scratch...
     while (@_) {
         my $arg = shift;
         if ( $known_action{$arg} ) {
@@ -282,7 +330,7 @@ sub parse_options {
             } elsif (ref $cond eq 'ARRAY') {
                 $self->restrict( @$cond );
             } else {
-                _croak( "-except argument must be regexp or array" )
+                _croak( "-only argument must be regexp or array" )
             };
         } else {
             _croak( "unknown option $arg" );
@@ -325,7 +373,9 @@ sub sigil_to_type {
 
 ### Command pattern split into prepare + execute
 
-# side effects + setup self->execute
+# input: none
+# output: none
+# side effects: modify target package + setup callback for self->execute
 sub prepare {
     my $self = shift;
 
@@ -356,6 +406,9 @@ sub prepare {
     };
 };
 
+# input: none
+# output: none
+# side effect: modify target package
 sub execute {
     my ($self) = @_;
 
@@ -369,12 +422,7 @@ sub execute {
 
 ### High-level effectful functions
 
-# Here and below, the following data format is used:
-# `symtable` := { `name` => { `type` => `$ref` } }
-# `type` is one of known variable types listed in @TYPES below
-# `$ref` is a reference of corresponding type
-
-# Don't touch NAME, PACKAGE, and GLOB that are alsoknown to Perl
+# Don't ever touch NAME, PACKAGE, and GLOB that are also known to Perl
 my @TYPES = qw(SCALAR ARRAY HASH CODE IO FORMAT);
 
 # In: symbol table hashref
@@ -422,10 +470,6 @@ sub replace_symbols {
     my $diff = $self->table_diff( $old_table, $table );
     return unless keys %$diff;
 
-    # apply change
-    $self->message( "package $self->{target} to be altered: ".dump_table($diff, $old_table) )
-        if DEBUG;
-
     $self->write_symbols( $diff );
 };
 
@@ -573,6 +617,12 @@ sub write_symbols {
 
     my $package = $self->{target};
 
+    if (DEBUG) {
+        my $old_table = $self->read_symbols;
+        $self->message( "package $self->{target} to be altered: ".dump_table($table, $old_table) )
+    };
+
+
     foreach my $name( keys %$table ) {
         my $copy = $table->{$name};
 
@@ -607,15 +657,37 @@ sub dump_table {
     foreach my $name( sort keys %$table ) {
         my $glob = $table->{$name};
         foreach my $type( sort keys %$glob ) {
-            push @out, "*$name\{$type\}=".($glob->{$type} || 'undef');
-            $out[-1] .= sprintf "[was: %s]", $old_table->{$name}{$type} || 'undef'
-                if $old_table;
+            push @out, "*$name\{$type\}=".(
+                $old_table
+                    ? _is_and_was( $glob->{$type}, $old_table->{$name}{$type} )
+                    : _ref2str( $glob->{$type} )
+            );
         };
     };
 
     return join ", ", @out;
 };
 
+sub _is_and_was {
+    my ($new, $old) = @_;
+
+    if ((refaddr $new || 0) != (refaddr $old || 0)) {
+        return _ref2str( $new )."[was: "._ref2str( $old )."]";
+    } else {
+        return _ref2str( $new )."[unchanged]";
+    };
+};
+
+# TODO find existing?
+sub _ref2str {
+    my $ref = shift;
+
+    return ref $ref
+        ? blessed $ref
+            ? sprintf "%s=%s(0x%x)", ref $ref, reftype $ref, refaddr $ref
+            : sprintf "%s(0x%x)", ref $ref, refaddr $ref
+        : 'undef';
+};
 
 sub message {
     my ($self, $msg) = @_;
@@ -682,10 +754,15 @@ L<http://search.cpan.org/dist/namespace-local/>
 
 =head1 SEE ALSO
 
-L<namespace::clean>, L<namespace::sweep>, L<namespace::autoclean>...
+L<namespace::clean> gave the inspiration for this module.
+
+L<namespace::sweep>, L<namespace::autoclean> and probably more also clean
+caller's namespace, but differently.
 
 L<B::Hooks::EndOfScope> is used as a backend.
 
+L<perlmod/Symbol Tables> explains how reading/writing the namespace works.
+
 =head1 LICENSE AND COPYRIGHT
 
 Copyright 2018 Konstantin S. Uvarin, C<< <khedin at gmail.com> >>
